// This file deals only with conflicts structure (set/get/add into this struct)
// This is used for MOVE agents and for created agents.
//   It does NOT solve the conflicts.

// NB: on IRIX, Solaris and Linux/glibc realloc() leaves unchanged the block if failure

#include <stdlib.h>
#include <stdio.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

#define INCCT 10

void initConflictTablePar (void)
{
  tsd_t *ptsd = getTSD();
  ptsd->nbConflicts = 0;
  ptsd->nbAllocConflicts = 0;
  ptsd->conflict = NULL;  // needed by realloc from addConflictStructurePar
}

void clearConflictTablePar (void)
{
  getTSD()->nbConflicts = 0;
}

void freeConflictTablePar (void)
{
  tsd_t *ptsd = getTSD();
  ptsd->nbConflicts = 0;
  ptsd->nbAllocConflicts = 0;
  if (ptsd->conflict)
    free (ptsd->conflict);
  ptsd->conflict = NULL;  // needed by realloc from addConflictStructurePar
}

/* ***************  MOVE conflicts  ********************* */

// return the id of the newly allocated structure or MEMERR
int addConflictStructurePar (void)
{
  tsd_t *ptsd = getTSD();
  if (ptsd->nbConflicts == ptsd->nbAllocConflicts){
    conflict_t *tmp;
    tmp = realloc (ptsd->conflict, (ptsd->nbAllocConflicts + INCCT) * sizeof(conflict_t));
    if (tmp == NULL)
      abcerror ("Cannot reallocate %d bytes in addConflictStructurePar.  Aborting\n", (ptsd->nbAllocConflicts + INCCT) * sizeof(conflict_t));
    ptsd->conflict = tmp;
    ptsd->nbAllocConflicts += INCCT;
  }
  ptsd->conflict[ptsd->nbConflicts].nbma = 0;
  ptsd->conflict[ptsd->nbConflicts].nbca = 0;
  return ptsd->nbConflicts ++;
}

// get number of conflicts in the conflict table
int getConflictsNumberPar (void)
{
  return getTSD()->nbConflicts;
}

void addConflictItemPar (int cfNumber, int item)
{
  int nb;
  tsd_t *ptsd = getTSD();
  I ((cfNumber >= 0) && (cfNumber < ptsd->nbConflicts));
  nb = ptsd->conflict[cfNumber].nbma;  // actual number of items in conflict
  I (nb < _nbDirs);          // max nbdir-1 items allowed in the structure
  ptsd->conflict[cfNumber].data[nb] = item;
  ptsd->conflict[cfNumber].nbma ++;
}

// get the number of items in a conflict structure
int getConflictItemsNumberPar (int cfNumber)
{
  I ((cfNumber >= 0) && (cfNumber < getTSD()->nbConflicts));
  return getTSD()->conflict[cfNumber].nbma;
}

// get the item number NO in the conflict structure number CFNUMBER
int getConflictDataPar (int cfNumber, int no)
{
  I ((cfNumber >= 0) && (cfNumber < getTSD()->nbConflicts));
  I ((no >= 0) && (no < getTSD()->conflict[cfNumber].nbma));
  return getTSD()->conflict[cfNumber].data[no];
}

// sort all the ids of this structure in increasing order
void sortConflictStructurePar (int cf)
{
  int *confdata, i,j, nb;
  I ((cf >= 0) && (cf < getTSD()->nbConflicts));
  confdata = getTSD()->conflict[cf].data;
  nb = getTSD()->conflict[cf].nbma;
  // bubble sort for the moment
  for (i=0 ; i<nb-1 ; i++)
    for (j=0 ; j<nb-1-i ; j++)
      if (confdata[j] > confdata[j+1]){
	int temp = confdata[j+1];
	confdata[j+1] = confdata[j];
	confdata[j] = temp;
      }
  for (i=0 ; i<nb-1 ; i++)
    I (confdata[i] <= confdata[i+1]);
}

/* ***************  created agents conflicts  ********************* */

void addConflictItemCAPar (int cfNumber, int item)
{
  int nb;
  tsd_t *ptsd = getTSD();
  I ((cfNumber >= 0) && (cfNumber < ptsd->nbConflicts));
  nb = ptsd->conflict[cfNumber].nbca;  // actual number of items in conflict
  if (nb < CFSIZE - _nbDirs){  // max CFSIZE - nbdirs items allowed in the structure
    ptsd->conflict[cfNumber].data[_nbDirs + nb] = item;
    ptsd->conflict[cfNumber].nbca ++;
  }else
    printf ("Agent creation refused, insufficient space to store it.  Continuing.\n");
}

// get the number of items in a conflict structure
int getConflictItemsNumberCAPar (int cfNumber)
{
  I ((cfNumber >= 0) && (cfNumber < getTSD()->nbConflicts));
  return getTSD()->conflict[cfNumber].nbca;
}

// get the item number NO in the conflict structure number CFNUMBER
int getConflictDataCAPar (int cfNumber, int no)
{
  I ((cfNumber >= 0) && (cfNumber < getTSD()->nbConflicts));
  I ((no >= 0) && (no < getTSD()->conflict[cfNumber].nbca));
  return getTSD()->conflict[cfNumber].data[_nbDirs + no];
}

// sort all the ids of this structure in increasing order (using the parent id)
void sortConflictStructureCAPar (int cf)
{
  int *confdata, i,j, nb;
  tsd_t *ptsd = getTSD();
  I ((cf >= 0) && (cf < ptsd->nbConflicts));
  confdata = &ptsd->conflict[cf].data[_nbDirs];
  nb = ptsd->conflict[cf].nbca;
  // bubble sort for the moment
  for (i=0 ; i<nb-1 ; i++)
    for (j=0 ; j<nb-1-i ; j++)
      // compare parents' id
      if (((agentinfo_t*)peekstructp(&ptsd->createdAgents, confdata[j], sizeof(agentinfo_t)))->parentid
	  > ((agentinfo_t*)peekstructp(&ptsd->createdAgents, confdata[j+1], sizeof(agentinfo_t)))->parentid){
	int temp = confdata[j+1];
	confdata[j+1] = confdata[j];
	confdata[j] = temp;
      }
  for (i=0 ; i<nb-1 ; i++)
    I (((agentinfo_t*)peekstructp(&ptsd->createdAgents, confdata[i], sizeof(agentinfo_t)))->parentid <= ((agentinfo_t*)peekstructp(&ptsd->createdAgents, confdata[i+1], sizeof(agentinfo_t)))->parentid);
}
