// This source file deals with the env struture

// EDGES -> walls at edges, TORUS -> use of modulo everywhere x, y
// normy/x were moved in intern.h, in order to be inlined or #defined

#include <stdlib.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

// type of a square (or of the table)
typedef struct{
  int type;              // 00-FREESQ, 01-RESSQ, 10-WALLSQ
  int state;             // 0-unoccupied, 1-occupied
  int typeGS;            // src type, if type == RESSQ
  int idGS;              // src id, if type == RESSQ
  int idAg;              // agent id, if an ag is in the square (state == OCCUP)
}square_t;

int _nbDirs;             // the number of directions used by agents, vision etc
int _dimy, _dimx;	 // env dim: 0.._dimy-1 and 0.._dimx-1
int _envType;	         // env type
int _diry[] = {-1, 1, 0, 0, -1, -1, 1, 1};
int _dirx[] = {0, 0, -1, 1, -1, 1, -1, 1};
static square_t *env;    // env table
static int nbObst;       // the number of obstacles in the environment


// set the number of directions (4 or 8)
// useful for vision and gradient propagation
void pmSetDirsNo (int n)
{
  I ((n == 4) || (n == 8));
  _nbDirs = n;
}

// ----------------------  tools  -------------------------------

// absolute->relative y
// -_dimy/2 <= returned value < _dimy/2
int pmtarCoordy (int y)
{
  return normy (y - getTSD()->crtAgenty + _dimy/2) - _dimy/2;
}

// absolute->relative x
// -_dimx/2 <= returned value < _dimx/2
int pmtarCoordx (int x)
{
  return normx (x - getTSD()->crtAgentx + _dimx/2) - _dimx/2;
}

// relative->absolute y
int pmtraCoordy (int ry)
{
  return normy (ry + getTSD()->crtAgenty);
}

// relative->absolute x
int pmtraCoordx (int rx)
{
  return normx (rx + getTSD()->crtAgentx);
}

// ---------------------  tools (cont)  ------------------------------

int pmtDir2y (int y)
{
  // the order of direction constants in agents.h matters
  I ((y>=0) && (y<_nbDirs));  // maybe I have to put from 0 to 8?
  return _diry[y];
}

int pmtDir2x (int x)
{
  // the order of direction constants in agents.h matters
  I ((x>=0) && (x<_nbDirs));
  return _dirx[x];
}

// ---------------------  tools (cont)  ------------------------------

// return 1 if an agent may move in (y,x)
int pmtaCanMove (int y, int x)
{
  ICHECKCOORDS (y, x);
  return (pmgaGetSquareType(y, x) != WALLSQ) && pmgaGetSquareState(y, x) == UNOCCUP;
}

int pmtmrCanMove (int ry, int rx)
{
  tsd_t *ptsd = getTSD();
  return pmtaCanMove(ry + ptsd->crtAgenty, rx + ptsd->crtAgentx);
}

// return 1 if the agent found in (y,x) may move in direction DIR, else 0
int pmtadCanMove (int y, int x, int dir)
{
  int newy, newx;
  ICHECKDIR (dir);
  newy = normy(y + _diry[dir]);
  newx = normx(x + _dirx[dir]);
  return pmtaCanMove(newy, newx);
}

int pmtdCanMove (int dir)
{
  tsd_t *ptsd = getTSD();
  return pmtadCanMove (ptsd->crtAgenty, ptsd->crtAgentx, dir);
}

// ------------------------------------------------------------

void pmCreateEnv (int dimY, int dimX, int type)
{
  int y, x;
  _dimy = dimY;
  _dimx = dimX;
  nbObst = 0;
  I((type == TORUS) || (type == EDGES));
  _envType = type;
  env = malloc(_dimx * _dimy * sizeof(square_t));
  if (env == NULL)
    abcerror ("Cannot allocate %d bytes in pmCreateEnv.  Aborting\n", _dimx * _dimy * sizeof(square_t));
  // initialize the environment
  for (y=0 ; y<_dimy ; y++)
    for (x=0 ; x<_dimx ; x++){
      env[y*_dimx + x].type = FREESQ;
      env[y*_dimx + x].state = UNOCCUP;
    }
  // if EDGES, put walls on edges
  if (_envType == EDGES){
    for (y=0 ; y<_dimy ; y++){
      pmPutObstacle (y, 0);
      pmPutObstacle (y, _dimx-1);
    }
    for (x=1 ; x<_dimx-1 ; x++){  // 0 and _dimx-1 are already used (see above)
      pmPutObstacle (0, x);
      pmPutObstacle (_dimy-1, x);
    }
  }
  initNewPos(dimY, dimX);
}

int pmgGetEnvDimy (void)
{
  return _dimy;
}

int pmgGetEnvDimx (void)
{
  return _dimx;
}

void pmPutObstacle (int y, int x)
{
  ICHECKCOORDS (y, x);
  I ((pmgaGetSquareType(y, x) == FREESQ) && (pmgaGetSquareState(y, x) == UNOCCUP));
  env[y*_dimx+x].type = WALLSQ;
  nbObst++;
}

void pmPutRandomObstacles (int nb)
{
  int y, x;
  // there are sufficiently free squares for all the obstacles
  I (nb + nbObst + pmgGetAgentNumber() <= _dimy * _dimx);  // agent + obst <= y*x
  I (nb + nbObst + getTotalResourceNumber() <= _dimy * _dimx);  // agent + gsrc <= y*x
  while (nb-- > 0){
    do{
      y = globrand(_dimy);
      x = globrand(_dimx);
    }while ((pmgaGetSquareType(y, x) != FREESQ) || (pmgaGetSquareState(y, x) == OCCUP));
    pmPutObstacle (y, x);
  }
}

// get the total number of obstacles
int getObstNumber (void)
{
  return nbObst;
}

// return FREESQ, RESSQ or WALLSQ
int pmgaGetSquareType (int y, int x)
{
  ICHECKCOORDS (y, x);
  return env[y*_dimx+x].type;
}

int pmlrGetSquareType (int ry, int rx)
{
  return pmgaGetSquareType (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmldGetSquareType (int dir)
{
  ICHECKDIR (dir);
  return pmlrGetSquareType (_diry[dir], _dirx[dir]);
}

int pmlmGetSquareType (void)
{
  tsd_t *ptsd = getTSD();
  return pmgaGetSquareType (ptsd->crtAgenty, ptsd->crtAgentx);
}

// return UNOCCUP or OCCUP
int pmgaGetSquareState (int y, int x)
{
  ICHECKCOORDS (y, x);
  return env[y*_dimx+x].state;
}

int pmlrGetSquareState (int ry, int rx)
{
  return pmgaGetSquareState (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmldGetSquareState (int dir)
{
  ICHECKDIR (dir);
  return pmlrGetSquareState (_diry[dir], _dirx[dir]);
}

int pmlmGetSquareState (void)
{
  tsd_t *ptsd = getTSD();
  return pmgaGetSquareState (ptsd->crtAgenty, ptsd->crtAgentx);
}

void setSquareState (int y, int x, int state)
{
  ICHECKCOORDS (y, x);
  I ((state == OCCUP) || (state == UNOCCUP));
  env[y*_dimx+x].state = state;
}

// put a GS in the square, set the square type and fill the appropriate information
void putSquareGS (int y, int x, int type, int id)
{
  ICHECKCOORDS (y, x);
  // assume the square is FREESQ and check type
  I (pmgaGetSquareType(y, x) == FREESQ);
  I ((type >= 0) && (type < getResourceTypesNumber()));
  env[y*_dimx+x].type = RESSQ;
  env[y*_dimx+x].typeGS = type;
  env[y*_dimx+x].idGS = id;
}


// return the type of the source in (y,x)
int getSquareSourceType (int y, int x)
{
  ICHECKCOORDS (y, x);
  // assume there is a gradient source in this square
  I (pmgaGetSquareType(y, x) == RESSQ);
  return env[y*_dimx+x].typeGS;
}

// return the usertype of the src in the square (y, x)
int pmgaGetSquareResType (int y, int x)
{
  return type2user(getSquareSourceType(y, x));
}

int pmlrGetSquareResType (int ry, int rx)
{
  return pmgaGetSquareResType (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmldGetSquareResType (int dir)
{
  ICHECKDIR (dir);
  return pmlrGetSquareResType (_diry[dir], _dirx[dir]);
}

int pmlmGetSquareResType (void)
{
  tsd_t *ptsd = getTSD();
  return pmgaGetSquareResType (ptsd->crtAgenty, ptsd->crtAgentx);
}

// return the load of the src in the square (y, x)
int pmgaGetSquareResLoad (int y, int x)
{
  ICHECKCOORDS (y, x);
  // assume there is a gradient source in this square
  I (pmgaGetSquareType(y, x) == RESSQ);
  return getResourceLoad(env[y*_dimx+x].typeGS, env[y*_dimx+x].idGS);
}

int pmlrGetSquareResLoad (int ry, int rx)
{
  return pmgaGetSquareResLoad (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmldGetSquareResLoad (int dir)
{
  ICHECKDIR (dir);
  return pmlrGetSquareResLoad (_diry[dir], _dirx[dir]);
}

int pmlmGetSquareResLoad (void)
{
  tsd_t *ptsd = getTSD();
  return pmgaGetSquareResLoad (ptsd->crtAgenty, ptsd->crtAgentx);
}

// return the id of the src in the square (y, x)
int getSquareSourceId (int y, int x)
{
  ICHECKCOORDS (y, x);
  // assume there is a gradient source in this square
  I (pmgaGetSquareType(y, x) == RESSQ);
  return env[y*_dimx+x].idGS;
}

int pmgaGetAgentId (int y, int x)
{
  ICHECKCOORDS (y, x);
  // assume there is an agent in this square
  I (pmgaGetSquareState(y, x) == OCCUP);
  return env[y*_dimx+x].idAg;
}

int pmlrGetAgentId (int ry, int rx)
{
  return pmgaGetAgentId (pmtraCoordy(ry), pmtraCoordx(rx));
}

int pmldGetAgentId (int dir)
{
  ICHECKDIR (dir);
  return pmlrGetAgentId (_diry[dir], _dirx[dir]);
}

// this doesn't set the square's state on OCCUP.  Maybe better to put it?
void setSquareAgentId (int y, int x, int id)
{
  ICHECKCOORDS (y, x);
  // assume there is NOT an agent in this square
  I (pmgaGetSquareState(y, x) == UNOCCUP);
  env[y*_dimx+x].idAg = id;
}

// called to free env
void freeEnv (void)
{
  if (env != NULL)
    free (env);
  freeNewPos ();
}
