// This file provides portable definitions for all non-portable calls,
//   like the threads for example.
// Created on Oct 22, 1999

// types

// thread id is passed as parameter
typedef void threadfunc_t (int);

// types internal to sproc.c, pthreads.c etc., not to the whole library
// maybe better to move it to another header, common to sproc.h, pthreads.h etc
typedef struct{
  threadfunc_t *func;
  int id;
}thdata_t;

// TODO: move this definition in conflicts.c
#define CFSIZE 16
typedef struct{
  int nbma;        // actual number of data in move conflict in this structure
  int nbca;        // actual number of data in ca conflict in this structure
  int data[CFSIZE];// id of the agents in conflict - 0..dirs-1 for move conflicts, and nbdirs..CFSIZE-1 for ca conflicts
}conflict_t;

// thread-specific data (only *one* key is used)
// all these fields are initialized in initSystemPar, except frontChanged
typedef struct{
  int id;             // thread id (0.._nbThreads-1)
  int cycle;          // the cycle

  int ytop, xtop;     // upper-left corner (inclusiv)
  int ybot, xbot;     // bottom-right corner (exclusiv), absolute coordinates

  int *frontChanged;  // bool, specify if the frontier changed
  int *nfrontier, *efrontier, *sfrontier, *wfrontier;  // boundary exchange method: here are copied the frontiers

  queue_t srcQueue;       // the queue (of 2 ints) used in updateGradients3

  int nbConflicts;        // actual number of conflicts
  int nbAllocConflicts;   // number of allocated conflicts
  conflict_t *conflict;   // the table with conflicts (move + created agents)

  stack_t createdAgents;  // the stack of agentinfo_t with local created agents

  int crtAgentId, crtAgenty, crtAgentx;  // info about the current agent
  unsigned int seed;

  char padding[256];      // padding to avoid L2 cache line false sharing
}tsd_t;


// -----------------------------------------------------------------

// functions

void createThreads (int, threadfunc_t *);  // NBTH-1 and itself execute FUNC
void syncThreads (void);

tsd_t *getTSD (void);

void lockfrch (void);
void unlockfrch (void);
void locknbAgent (void);
void unlocknbAgent (void);
void lockAgent (void);
void unlockAgent (void);
void lockCAgent (void);
void unlockCAgent (void);
void lockDAgent (void);
void unlockDAgent (void);
void lockKAgent (void);
void unlockKAgent (void);
void lockMAgent (void);
void unlockMAgent (void);
void lockEmptyAgent (void);
void unlockEmptyAgent (void);
void lockRandomAgent (void);
void unlockRandomAgent (void);
void lockEmptyResN (int);
void unlockEmptyResN (int);
void lockUnvisResN (int);
void unlockUnvisResN (int);
void lockTotalResL (int);
void unlockTotalResL (int);
