// This file deals with the newpos structure.
// This is used for MOVE agents and for created agents

#include <stdlib.h>
#include <string.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

#define ZEROAG 0  // no agent in conflict
#define ONEMA  1  // one move agent here
#define ONECA  2  // one created agent here
#define MANYAG 3  // there is a conflict here (id is the conflict index)

// move and created agents use the same conflict table
typedef struct{
  int nb;     // type of id (see above)
  int id;     // id of the agent or id of the conflict element
}newpos_t;
static newpos_t *newpos;

// NB: you have to clear it before use
void initNewPos (int ysize, int xsize)
{
  newpos = malloc (ysize * xsize * sizeof(newpos_t));
  if (newpos == NULL)
    abcerror ("Cannot allocate %d bytes in initNewPos.  Aborting\n", ysize * xsize * sizeof(newpos_t));
}

// clear the table
void clearNewPos (void)
{
  tsd_t *ptsd = getTSD();
  int ytop = ptsd->ytop, ybot = ptsd->ybot, xtop = ptsd->xtop, xbot = ptsd->xbot;
  int y;
  for (y=ytop ; y<ybot ; y++)
    memset (newpos+y*_dimx+xtop, ZEROAG, (xbot-xtop)*sizeof(newpos_t));
}

// add a new position.  If in the square there is already an obj,
//   then a conflict item is used.
static void addnp (int y, int x, int id, int type)
{
  int tmpitem;
  int elem = y*_dimx+x;
  void (*addcfitem)(int, int);
  ICHECKCOORDS (y, x);
  I ((type == ONEMA) || (type == ONECA));
  if (type == ONEMA)
    addcfitem = addConflictItemPar;
  else
    addcfitem = addConflictItemCAPar;
  switch (newpos[elem].nb){
  case ZEROAG:
    newpos[elem].id = id;
    newpos[elem].nb = type;  // there is an agent here
    break;
  case ONEMA:
  case ONECA:
    tmpitem = addConflictStructurePar();
    I (tmpitem >= 0);
    if (newpos[elem].nb == ONEMA)  // it was a move agent
      addConflictItemPar (tmpitem, newpos[elem].id);
    else  // it was a created agent
      addConflictItemCAPar (tmpitem, newpos[elem].id);
    addcfitem (tmpitem, id);
    newpos[elem].id = tmpitem;
    newpos[elem].nb = MANYAG;  // there is a conflict here
    break;
  case MANYAG:
    addcfitem (newpos[elem].id, id);
    break;
  default:
    I (0);  // cannot arrive here
  }
}

void addNewPos (int y, int x, int id)
{
  addnp (y, x, id, ONEMA);
}

void addNewPosCA (int y, int x, int id)
{
  addnp (y, x, id, ONECA);
}

void freeNewPos (void)
{
  if (newpos != NULL)
    free (newpos);
}
