// File which deals with queue data structure of two integers for the moment
////   and structures (not the pointer, but the structure is copied)
// This is NOT re-entrant.  You have to use lock functions before and after the call if in parallel
// You have to call clearQueue() before starting to use it (the queue is not circular)

#include <stdlib.h>
#include "agents.h"
#include "intern.h"
#include "nana.h"

// implemented as an array allocated with malloc (growing possible with resetQ)
// NB: nbelem is >= TOTAL number of elems added in the queue (addIndex is not used with modulo, so it does not restart with 0 if it has arrived to lastIndex)
// the algorithm is thus optimised
void initQueue (queue_t *q, int nbelem)
{
  q->lastIndex = nbelem - 1;
  q->addIndex = q->removeIndex = 0;
  q->data = malloc (nbelem * sizeof(int));
  if (q->data == NULL)
    abcerror("Cannot allocate %d bytes in initQueue.  Aborting.\n", nbelem * sizeof(int));
}

// NOT re-entrant
// this does NOT have to become greater than lastIndex (the queue is not circular)
void addints (queue_t *q, int e1, int e2)
{
  I (q);
  I (q->addIndex <= q->lastIndex - 2);
  q->data[q->addIndex++] = e1;
  q->data[q->addIndex++] = e2;
}

// NOT re-entrant
// NB: the queue is not circular
void removeints (queue_t *q, int *p1, int *p2)
{
  I (q && p1 && p2);
  I (q->removeIndex <= q->addIndex - 2);
  I (q->data);
  *p1 = q->data[q->removeIndex++];
  *p2 = q->data[q->removeIndex++];
}

int depthQueue (queue_t *q)
{
  I (q);
  return (q->addIndex - q->removeIndex);
}

void resetQueue (queue_t *q, int nbelem)
{
  I (q && q->data);
  q->addIndex = q->removeIndex = 0;
  if (nbelem - 1 > q->lastIndex){
    // grow the queue data
    VL (("! A queue grows up!  The perfs of gradsrc may decrease\n"));
    q->data = realloc (q->data, nbelem * sizeof(int));
    if (q->data == NULL)
      abcerror("Cannot reallocate %d bytes in resetQueue.  Aborting.\n", nbelem * sizeof(int));
    q->lastIndex = nbelem - 1;
  }
}

void endQueue (queue_t *q)
{
  I (q);
  I (q->data);
  // if this check is not good, use resetQueue() before
  I (q->addIndex == q->removeIndex);
  free (q->data);
}
