// This source file deals with Changes files.
// Always save and use first y, then x

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

#define ENVEXT    ".obs"  // extensions of the Changes files
#define AGEXT     ".ags"
#define RESEXT    ".res"
#define MAXEXT 4          // maximum length of the extensions above

static char changesBasename[FILENAME_MAX] = "Changes";
static FILE *envFile, *resFile, *agentFile;
static int isCreated;  // are the Changes files created?

// fill fn and return the value of isCreated
int getChangesFileName (char *fn)
{
  strcpy (fn, changesBasename);
  return isCreated;
}

static FILE *createFile (const char *basename, char *ext)
{
  char filename[FILENAME_MAX];
  sprintf (filename, "%s%s", basename, ext);
  return fopen(filename, "w");
}

//create Changes files
static int createFiles (char *basename)
{
  I (strlen(basename) < FILENAME_MAX-MAXEXT);
  // env file creation
  envFile = createFile(basename, ENVEXT);
  if (envFile == NULL)
    return FILEERR;
  // res file creation
  resFile = createFile(basename, RESEXT);
  if (resFile == NULL)
    return FILEERR;
  // agent file creation
  agentFile = createFile(basename, AGEXT);
  if (agentFile == NULL)
    return FILEERR;
  return OK;
}

// write headers of Changes files
static void writeHeaders (void)
{
  I (envFile && agentFile && resFile);
  // writes the top line of env file, ie. the environment dimensions
  fprintf(envFile, "%d %d\n", _dimy, _dimx);
  fprintf(resFile, "%d %d\n", _dimy, _dimx);
  fprintf(agentFile, "%d %d\n", _dimy, _dimx);
}

static void writeObstacles (void)
{
  int y, x;
  int cycle = getTSD()->cycle;
  I(envFile);
  for (y=0 ; y<_dimy ; y++)
    for (x=0 ; x<_dimx ; x++)
      if (pmgaGetSquareType(y, x) == WALLSQ)
	fprintf(envFile, "W %4d %4d %4s %4d\n", y, x, "", cycle);
}

static void writeResources (void)
{
  int type, i;
  int cycle = getTSD()->cycle;
  I (resFile);
  for (type=0 ; type<getResourceTypesNumber() ; type++){
    int usertype = type2user (type);
    for (i=0 ; i<getResourceNumber(type) ; i++){
      int load = getResourceLoad (type, i);
      int y = getResourcey (type, i);
      int x = getResourcex (type, i);
      fprintf(resFile, "R %4d %4d %4d %4d %4d\n", usertype, y, x, load, cycle);
    }
  }
}

static void writeAgents (void)
{
  int ag, cycle = getTSD()->cycle;
  for (ag=0 ; ag<getAgentNumbermax() ; ag++)
    if (isAlive(ag))
      fprintf(agentFile,"A %4d %4d %4d %4d %4d\n",
	      ag, pmgiGetAgenty(ag), pmgiGetAgentx(ag), pmgiGetAgentLoad(ag), cycle);
}

// provide the basename of the files to use with pmSaveChanges()
// or to cut the name to FILENAME_MAX and eliminate the return value ?
int pmInitChanges (char *basename)
{
  I (!isCreated);
  if (strlen(basename) >= FILENAME_MAX-MAXEXT)
    return ERR;
  strcpy (changesBasename, basename);
  isCreated = 0;
  return OK;
}

int initChangesAppend (char *basename)
{
  if (strlen(basename) >= FILENAME_MAX-MAXEXT)
    return ERR;
  strcpy (changesBasename, basename);
  isCreated = 1;
  return OK;
}

// user function
// save changes of the system, which may be read with the visualisation program
// filename is used only the first time, when it creates the file
int pmSaveChanges (void)
{
  if (! isCreated){
    int ret = createFiles(changesBasename);
    if (ret != OK)
      return ret;
    isCreated = 1;
    writeHeaders ();
    writeObstacles ();  // obstacles are static, write them only the 1st cycle
  }
  writeResources ();
  writeAgents ();
  return OK;
}

// internal library function
void closeChangesFiles (void)
{
  if (envFile != NULL)
    fclose (envFile);
  if (resFile != NULL)
    fclose (resFile);
  if (agentFile != NULL)
    fclose (agentFile);
  isCreated = 0;
}
