// This file deals with the statistic file: init-ing, writing and ending

#include <stdio.h>
#include <string.h>
#include "agents.h"    // for typechecking only
#include "intern.h"    // for typechecking only
#include "lowinterface.h"
#include "nana.h"

char statfilename[FILENAME_MAX] = "Statistics";
static FILE *statfile;
static int isOpen;     // 1 if the file is opened, else 0
static int isWritten;  // the top line of the file is written
static char mode[2] = "w";  // the mode of opening the statistics file

// create or truncate the stat file FN
int pmInitStat (char *fn)
{
  I (!isOpen);
  if (strlen(fn) >= FILENAME_MAX)
    return ERR;
  strcpy (statfilename, fn);
  strcpy (mode, "w");
  isWritten = 0;
  return OK;
}

// open existing file FN
int initStatAppend (char *fn)
{
  I (!isOpen);
  if (strlen(fn) >= FILENAME_MAX)
    return ERR;
  strcpy (statfilename, fn);
  strcpy (mode, "a");
  isWritten = 1;
  return OK;
}

// save all the information in the stat file
void pmSaveStat (void)
{
  int i;
  if (!isOpen){
    // open the statistics file
    statfile = fopen (statfilename, mode);
    I (statfile);
    isOpen = 1;
  }
  if (!isWritten){
    // write the top line of the statistic file
    fprintf (statfile, "#cycle");
    for (i=0 ; i<getResourceTypesNumber() ; i++)
      if (getResourceNumber(i) != 0)
	fprintf (statfile, "\tload%d", type2user(i));
/*     fprintf (statfile, "\tminAg"); */
    for (i=0 ; i<getResourceTypesNumber() ; i++)
      if (getResourceNumber(i) != 0)
	fprintf (statfile, "\tempty%d", type2user(i));
    for (i=0 ; i<getResourceTypesNumber() ; i++)
      if (getResourceNumber(i) != 0)
	fprintf (statfile, "\tinit%d", type2user(i));
    fprintf (statfile, "\tnbAg");
    fprintf (statfile, "\trandAg");
    fprintf (statfile, "\temptyAg\n");
    isWritten = 1;
  }

  // write the information
  // cycle
  fprintf (statfile, "%7d", getTSD()->cycle);
  // for each type of resource, the percentage of objects in resource
  // minX = (loadSourceX / loadInitSources)
  for (i=0 ; i<getResourceTypesNumber() ; i++)
    if (getResourceNumber(i) != 0)
      fprintf (statfile, "\t%6.2f", (getResTotalLoad(i)*100.0) / getInitResLoad());
  // minAg = (total agent load / loadInitSources)
/*   fprintf (statfile, "\tminAg"); */
  // for each type of resource, the percentage of empty resources
  // emptyX = (number of empty sources / total number of sources)
  for (i=0 ; i<getResourceTypesNumber() ; i++)
    if (getResourceNumber(i) != 0)
      fprintf (statfile, "\t%6.2f", (getEmptyResNumber(i)*100.0) / getResourceNumber(i));
  // for each type of resource, the percentage of unvisited resources
  // initX = (number of unvisited sources / total number of sources)
  for (i=0 ; i<getResourceTypesNumber() ; i++)
    if (getResourceNumber(i) != 0)
      fprintf (statfile, "\t%6.2f", (getUnvisitedResNumber(i)*100.0) / getResourceNumber(i));
  // the number of agents during this cycle
  // nbAg = number of agents
  fprintf (statfile, "\t%6d", pmgGetAgentNumber());
  // the number of agents doing random movement during this cycle
  // randAg = (number of random robots / total number of robots)
  fprintf (statfile, "\t%6.2f", (100.0*getRandomAgentNumber()) / pmgGetAgentNumber());
  // the number of empty agents during this cycle
  // emptyAg = (number of empty robots / total number of robots)
  fprintf (statfile, "\t%6.2f\n", (100.0*getEmptyAgentNumber()) / pmgGetAgentNumber());
}

// close the stat file
void endStat (void)
{
  if (isOpen){
    fclose (statfile);
    isOpen = 0;
  }
  isWritten = 0;
}

// get the statistic file name
// return 0 if statistic file not used (not opened), and 1 if opened (and put
//   stat file name in FILENAME)
int getStatFileName (char *filename)
{
  strcpy (filename, statfilename);
  return isOpen;
}
