// This file provides functions to deal with sproc library of SGI
// TODO: write a common interface

#include <stdlib.h>
#include <ulocks.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include "agents.h"
#include "intern.h"
#include "lowinterface.h"
#include "nana.h"

static thdata_t *data;
static usptr_t *pArena;
static barrier_t *pBarrier;
static pid_t *th_t;
static tsd_t **tsd;  // Thread Specific Data (emulation by index)
static char *tmpFileName;
static ulock_t mutfrch;
static ulock_t mutMAg;
static ulock_t mutDAg;
static ulock_t mutKAg;
static ulock_t mutCAg;
static ulock_t mutAg;
static ulock_t mutnbAg;
static ulock_t mutEmptyAg;
static ulock_t mutRandomAg;
static ulock_t *mutEmptyResN;
static ulock_t *mutUnvisResN;
static ulock_t *mutTotalResL;

void lockTotalResL (int type)
{
  I (type <= getResourceTypesNumber());
  if (_nbThreads > 1){
    int ret = ussetlock (mutTotalResL[type]);
    if (ret != 1){
      printf ("lockTotalResL: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockTotalResL (int type)
{
  I (type <= getResourceTypesNumber());
  if (_nbThreads > 1){
    int ret = usunsetlock (mutTotalResL[type]);
    if (ret != 0){
      printf ("unlockTotalResL: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockEmptyResN (int type)
{
  I (type <= getResourceTypesNumber());
  if (_nbThreads > 1){
    int ret = ussetlock (mutEmptyResN[type]);
    if (ret != 1){
      printf ("lockEmptyResN: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockEmptyResN (int type)
{
  I (type <= getResourceTypesNumber());
  if (_nbThreads > 1){
    int ret = usunsetlock (mutEmptyResN[type]);
    if (ret != 0){
      printf ("unlockEmptyResN: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockUnvisResN (int type)
{
  I (type <= getResourceTypesNumber());
  if (_nbThreads > 1){
    int ret = ussetlock (mutUnvisResN[type]);
    if (ret != 1){
      printf ("lockUnvisResN: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockUnvisResN (int type)
{
  I (type <= getResourceTypesNumber());
  if (_nbThreads > 1){
    int ret = usunsetlock (mutUnvisResN[type]);
    if (ret != 0){
      printf ("unlockUnvisResN: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockEmptyAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutEmptyAg);
    if (ret != 1){
      printf ("lockEmptyAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockEmptyAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutEmptyAg);
    if (ret != 0){
      printf ("unlockEmptyAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockRandomAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutRandomAg);
    if (ret != 1){
      printf ("lockRandomAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockRandomAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutRandomAg);
    if (ret != 0){
      printf ("unlockRandomAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockMAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutMAg);
    if (ret != 1){
      printf ("lockMAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockMAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutMAg);
    if (ret != 0){
      printf ("unlockMAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockCAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutCAg);
    if (ret != 1){
      printf ("lockCAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockCAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutCAg);
    if (ret != 0){
      printf ("unlockCAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockDAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutDAg);
    if (ret != 1){
      printf ("lockDAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockDAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutDAg);
    if (ret != 0){
      printf ("unlockDAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockKAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutKAg);
    if (ret != 1){
      printf ("lockKAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockKAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutKAg);
    if (ret != 0){
      printf ("unlockKAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutAg);
    if (ret != 1){
      printf ("lockAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutAg);
    if (ret != 0){
      printf ("unlockAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void locknbAgent (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutnbAg);
    if (ret != 1){
      printf ("locknbAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlocknbAgent (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutnbAg);
    if (ret != 0){
      printf ("unlocknbAgent: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void lockfrch (void)
{
  if (_nbThreads > 1){
    int ret = ussetlock (mutfrch);
    if (ret != 1){
      printf ("lockfrch: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

void unlockfrch (void)
{
  if (_nbThreads > 1){
    int ret = usunsetlock (mutfrch);
    if (ret != 0){
      printf ("unlockfrch: error %d, errno=%d.  Aborting\n", ret, errno);
      abort ();
    }
  }
}

// -----------------------------------------------------------------------

// simulate the TSD
static int me (void)
{
  int i;
  pid_t m = getpid ();
  for (i=0 ; i<_nbThreads ; i++)
    if (th_t[i] == m)
      return i;
  I (0);  // it have not to arrive here
  return -1;  // only for not having warning
}

static void setSpecific (void *p)
{
  tsd[me()] = p;
}

static void *getSpecific (void)
{
  return tsd[me()];
}

tsd_t *getTSD (void)
{
  return getSpecific();
}

// wrap the thread function
static void thfunc (void *pv)
{
  tsd_t *pdata;
  thdata_t *p = (thdata_t*)pv;
  syncThreads ();  // needed?
  // allocate the thread-specific data
  pdata = malloc (sizeof (tsd_t));
  if (pdata == NULL)
    abcerror ("Cannot allocate %d bytes in thfunc.  Aborting.\n", sizeof (tsd_t));
  setSpecific (pdata);
  p->func (p->id);
  free (pdata);
}

static void waitThreads (void)
{
  int i;
  for (i=1 ; i<_nbThreads ; i++)  // wait for the threads
    wait (NULL);
  for (i=getResourceTypesNumber()-1 ; i>=0 ; i--){ // in inverse order (even if not necessary)
    usfreelock (mutUnvisResN[i], pArena);
    usfreelock (mutEmptyResN[i], pArena);
    usfreelock (mutTotalResL[i], pArena);
  }
  free (mutTotalResL);
  free (mutEmptyResN);
  free (mutUnvisResN);
  usfreelock (mutRandomAg, pArena);
  usfreelock (mutEmptyAg, pArena);
  usfreelock (mutMAg, pArena);
  usfreelock (mutCAg, pArena);
  usfreelock (mutDAg, pArena);
  usfreelock (mutKAg, pArena);
  usfreelock (mutAg, pArena);
  usfreelock (mutnbAg, pArena);
  usfreelock (mutfrch, pArena);
  free_barrier (pBarrier);
  unlink (tmpFileName);
  free (tsd);
  free (data);
  free (th_t);
}

void createThreads (int nbth, threadfunc_t *func)
{
  int i;
  I (nbth >= 0);
  data = malloc (nbth * sizeof (thdata_t));
  th_t = malloc (nbth * sizeof (pid_t));
  if ((data == NULL) || (th_t == NULL))
    abcerror ("Cannot allocate %d bytes twice in createThreads.  Aborting.\n", nbth * sizeof (thdata_t));
  tsd = malloc (nbth * sizeof (tsd_t*));  // here because it is global to all the threads
  if (tsd == NULL)
    abcerror ("Cannot allocate %d bytes in createThreads.  Aborting.\n", nbth * sizeof (tsd_t*));
  for (i=0 ; i<nbth ; i++){
    data[i].func = func;
    data[i].id = i;
  }
  // system initialization
  if (usconfig(CONF_INITUSERS, nbth) == -1){
    printf ("createThreads: error usconfig.  Aborting\n");
    abort ();
  }
  if (usconfig(CONF_INITSIZE, 1048576) == -1){  // 1MB
    printf ("createThreads: error usconfig.  Aborting\n");
    abort ();
  }
  // usconfig (CONF_INITSIZE) if I want to use usmalloc instead of malloc
  prctl(PR_SETSTACKSIZE, (rlim_t)100000);  // to have memory for >2 threads
  tmpFileName = tmpnam(NULL);
  if (tmpFileName == NULL){
    printf ("createThreads: error in tmpnam, errno=%d.  Aborting\n", errno);
    abort ();
  }
  VL (("Temporary file name: %s\n", tmpFileName));
  pArena = usinit(tmpFileName);
  if (pArena == NULL){
    printf ("createThreads: error in usinit, errno=%d.  Aborting\n", errno);
    abort ();
  }
  pBarrier = new_barrier (pArena);
  if (pBarrier == NULL){
    printf ("createThreads: error in new_barrier, errno=%d.  Aborting\n", errno);
    abort ();
  }
  // initialize the mutexes
  mutfrch = usnewlock (pArena);
  if (mutfrch == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutnbAg = usnewlock (pArena);
  if (mutnbAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutAg = usnewlock (pArena);
  if (mutAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutDAg = usnewlock (pArena);
  if (mutDAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutKAg = usnewlock (pArena);
  if (mutKAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutCAg = usnewlock (pArena);
  if (mutCAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutMAg = usnewlock (pArena);
  if (mutMAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutEmptyAg = usnewlock (pArena);
  if (mutEmptyAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutRandomAg = usnewlock (pArena);
  if (mutRandomAg == NULL){
    printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
    abort ();
  }
  mutTotalResL = malloc (getResourceTypesNumber() * sizeof(ulock_t));
  mutEmptyResN = malloc (getResourceTypesNumber() * sizeof(ulock_t));
  mutUnvisResN = malloc (getResourceTypesNumber() * sizeof(ulock_t));
  if (!mutTotalResL || !mutEmptyResN || !mutUnvisResN)
    abcerror ("Cannot allocate %d bytes three times in createThreads.  Aborting.\n", getResourceTypesNumber() * sizeof(ulock_t));
  for (i=0 ; i<getResourceTypesNumber() ; i++){
    mutTotalResL[i] = usnewlock (pArena);
    if (mutTotalResL[i] == NULL){
      printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
      abort ();
    }
    mutEmptyResN[i] = usnewlock (pArena);
    if (mutEmptyResN[i] == NULL){
      printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
      abort ();
    }
    mutUnvisResN[i] = usnewlock (pArena);
    if (mutUnvisResN[i] == NULL){
      printf ("createThreads: error in usnewlock, errno=%d.  Aborting\n", errno);
      abort ();
    }
  }
  // _nbThreads must be set before starting any thread (used by waitThreads and thfunc)
  _nbThreads = nbth;
  // threads creation and parallel execution
  for (i=1 ; i<nbth ; i++){
    int pid = sproc(thfunc, PR_SADDR, &data[i]);
    if (pid == -1){
      printf ("createThreads: error %d creating thread %d, errno=%d.  Aborting\n",
	      pid, i, errno);
      perror ("perror");
      exit (1);
    }
    th_t[i] = pid;
  }
  th_t[0] = getpid ();
  thfunc (&data[0]);
  waitThreads ();  // wait for all the other threads
}

void syncThreads (void)
{
  barrier (pBarrier, _nbThreads);
}
