// File which deals with stack data structure of integers
//   and structures (not the pointer, but the structure is copied)
// This is NOT re-entrant.  You have to use lock functions before and after the call

#include <string.h>
#include <stdlib.h>
#include "agents.h"
#include "intern.h"
#include "nana.h"

/* **********************  generic stack  ********************* */

// implemented as an array (re)allocated with malloc/realloc
void initStack (stack_t *s, int inc)
{
  s->nbAllocElem = 0;
  s->nbElem = 0;
  s->increment = inc;
  s->data = NULL;  // no data for the moment
}

int depthStack (stack_t *s)
{
  I (s);
  return s->nbElem;
}

void clearStack (stack_t *s)
{
  I (s);
  s->nbElem = 0;
}

void endStack (stack_t *s)
{
  I (s);
  // if this check is not good, use clearStack() before
  I (s->nbElem == 0);
  if (s->data)
    free (s->data);
}

/* ***********************  stack of ints   ************************* */

// NOT re-entrant
void pushint (stack_t *s, int e)
{
  I (s);
  I (s->nbAllocElem >= s->nbElem);
  if (s->nbElem == s->nbAllocElem){  // need to increase the data
    void *tmp = realloc (s->data, (s->nbAllocElem + s->increment) * sizeof(int));
    if (tmp == NULL)
      abcerror("Cannot reallocate %d bytes in pushint.  Aborting.\n", (s->nbAllocElem + s->increment) * sizeof(int));
    s->data = tmp;
    s->nbAllocElem += s->increment;
  }
  ((int*)s->data)[s->nbElem++] = e;
}

// NOT re-entrant
// this implementation does not deallocate memory
int popint (stack_t *s)
{
  I (s);
  I (s->nbElem > 0);
  I (s->data);
  return ((int*)s->data)[--s->nbElem];
}

// get the element number INDEX (like a vector)
int peekint (stack_t *s, int ind)
{
  I (s);
  I (ind >= 0);
  I (s->nbElem > ind);
  I (s->data);
  return ((int*)s->data)[ind];
}

/* ***********************  stack of structures   ************************* */

// NOT re-entrant
// size = size of an element
// return the index of the new element
// the structure, not the pointer is copied
int pushstruct (stack_t *s, void *pe, int size)
{
  I (s);
  I (s->nbAllocElem >= s->nbElem);
  if (s->nbElem == s->nbAllocElem){  // need to increase the data
    void *tmp = realloc (s->data, (s->nbAllocElem + s->increment) * size);
    if (tmp == NULL)
      abcerror("Cannot reallocate %d bytes in pushstruct.  Aborting.\n", (s->nbAllocElem + s->increment) * size);
    s->data = tmp;
    s->nbAllocElem += s->increment;
  }
  memcpy ((char*)s->data + s->nbElem * size, pe, size);
  return s->nbElem++;
}

// NOT re-entrant
// this implementation does not deallocate memory
void popstruct (stack_t *s, void *pe, int size)
{
  I (s);
  I (s->nbElem > 0);
  I (s->data);
  memcpy (pe, (char*)s->data + (--s->nbElem) * size, size);
}

// get the element number INDEX (like a vector)
void peekstruct (stack_t *s, int ind, void *pe, int size)
{
  I (s);
  I (ind >= 0);
  I (s->nbElem > ind);
  I (s->data);
  memcpy (pe, (char*)s->data + ind * size, size);
}

// get a pointer to the element number INDEX (like a vector)
void *peekstructp (stack_t *s, int ind, int size)
{
  I (s);
  I (ind >= 0);
  I (s->nbElem > ind);
  I (s->data);
  return (void*)((char*)s->data + ind * size);
}
